/*
 * Copyright (c) 2010-2012, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.oracle.xyzzyx.jdi;

import com.sun.jdi.ClassNotLoadedException;
import com.sun.jdi.Field;
import com.sun.jdi.IncompatibleThreadStateException;
import com.sun.jdi.InvalidTypeException;
import com.sun.jdi.InvocationException;
import com.sun.jdi.Method;
import com.sun.jdi.ObjectReference;
import com.sun.jdi.ThreadReference;
import com.sun.jdi.ReferenceType;
import com.sun.jdi.Value;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 *
 */
public class XyzzyxJDIObjectReference extends XyzzyxJDIValue implements ObjectReference {
    public XyzzyxJDIObjectReference(XyzzyxJDIVirtualMachine xyzzyxvm, ObjectReference underlying) {
        super(xyzzyxvm, underlying);
    }

    @Override
    public List<ObjectReference> referringObjects(long count) {
        return XyzzyxJDIWrapper.wrapObjectReferences(virtualMachine(), underlying().referringObjects(count));
    }

    @Override
    public void disableCollection() {
        underlying().disableCollection();
    }

    @Override
    public void enableCollection() {
        underlying().enableCollection();
    }

    @Override
    public int entryCount() throws IncompatibleThreadStateException {
        return underlying().entryCount();
    }

    @Override
    public Value getValue(Field field) {
        return XyzzyxJDIWrapper.wrap(virtualMachine(), underlying().getValue(XyzzyxJDIWrapper.unwrap(field)));
    }

    @Override
    public Map<Field, Value> getValues(List<? extends Field> wrappedFields) {
        virtualMachine().setLastFieldAccessException(null);

        // We will find fields which have no getters, and call the underlying
        // getValues to get values for all of them in one fell swoop.
        Map<Field, Field> unwrappedToWrappedMap = new HashMap<>();
        List<Field> noGetterUnwrappedFields = new ArrayList<>();    // fields that don't have getters

        // But first, for fields that do have getters, call invokeMethod
        // or we will call XyzzyxJDIGetValue for each, depending on doInvokes
        Map<Field, Value> result = new HashMap<>();
        XyzzyxJDIReferenceType wrappedClass = (XyzzyxJDIReferenceType)referenceType();
        ReferenceType unwrappedClass = XyzzyxJDIWrapper.unwrap(wrappedClass);

        // Create the above Maps and lists
        for (Field wrappedField : wrappedFields) {
            Field unwrapped = XyzzyxJDIWrapper.unwrap(wrappedField);
            unwrappedToWrappedMap.put(unwrapped, wrappedField);
            noGetterUnwrappedFields.add(unwrapped);
        }

        // Get values for all the noGetter fields.  Note that this gets them in a single JDWP trip
        Map<Field, Value> unwrappedFieldValues = underlying().getValues(noGetterUnwrappedFields);

        // for each input Field, create a result map entry with that field as the
        // key, and the value returned by getValues, or null if the field is invalid.

        // Make a pass over the unwrapped no getter fields and for each, put its
        // wrapped version, and wrapped value into the result Map.
        for (Map.Entry<Field, Field> unwrappedEntry: unwrappedToWrappedMap.entrySet()) {
            Field wrappedField = unwrappedEntry.getValue();
            Value resultValue = XyzzyxJDIWrapper.wrap(virtualMachine(),
                                             unwrappedFieldValues.get(unwrappedEntry.getKey()));
            result.put(wrappedField, resultValue);
        }
        return result;
    }

    @Override
    public XyzzyxJDIValue invokeMethod(ThreadReference thread, Method method, List<? extends Value> values, int options)
            throws InvalidTypeException, ClassNotLoadedException, IncompatibleThreadStateException, InvocationException {
        Value value =
                underlying().invokeMethod(
                    XyzzyxJDIWrapper.unwrap(thread), XyzzyxJDIWrapper.unwrap(method),
                    XyzzyxJDIWrapper.unwrapValues(values), options);
        return XyzzyxJDIWrapper.wrap(virtualMachine(), value);
    }

    @Override
    public boolean isCollected() {
        return underlying().isCollected();
    }

    @Override
    public XyzzyxJDIThreadReference owningThread() throws IncompatibleThreadStateException {
        return XyzzyxJDIWrapper.wrap(virtualMachine(), underlying().owningThread());
    }

    @Override
    public XyzzyxJDIReferenceType referenceType() {
        return XyzzyxJDIWrapper.wrap(virtualMachine(), underlying().referenceType());
    }

    @Override
    public void setValue(Field field, Value value) throws InvalidTypeException, ClassNotLoadedException {
        underlying().setValue(XyzzyxJDIWrapper.unwrap(field), XyzzyxJDIWrapper.unwrap(value));
    }

    @Override
    public long uniqueID() {
        return underlying().uniqueID();
    }

    @Override
    public List<ThreadReference> waitingThreads() throws IncompatibleThreadStateException {
        return XyzzyxJDIWrapper.wrapThreads(virtualMachine(), underlying().waitingThreads());
    }

    @Override
    protected ObjectReference underlying() {
        return (ObjectReference) super.underlying();
    }
}
