/*
 * Copyright (c) 2010-2012, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.oracle.xyzzyx.jdi;

import com.sun.jdi.AbsentInformationException;
import com.sun.jdi.ClassLoaderReference;
import com.sun.jdi.ClassObjectReference;
import com.sun.jdi.Field;
import com.sun.jdi.Location;
import com.sun.jdi.Method;
import com.sun.jdi.ObjectReference;
import com.sun.jdi.ReferenceType;
import com.sun.jdi.Value;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 *
 */
public class XyzzyxJDIReferenceType extends XyzzyxJDIType implements ReferenceType {
    public XyzzyxJDIReferenceType(XyzzyxJDIVirtualMachine xyzzyxvm, ReferenceType underlying) {
        super(xyzzyxvm, underlying);
    }

    @Override
    public byte[] constantPool() {
        return underlying().constantPool();
    }

    @Override
    public int constantPoolCount() {
        return underlying().constantPoolCount();
    }

    @Override
    public List<ObjectReference> instances(long count) {
        return XyzzyxJDIWrapper.wrapObjectReferences(virtualMachine(), underlying().instances(count));
    }

    @Override
    public int majorVersion() {
        return underlying().majorVersion();
    }

    @Override
    public int minorVersion() {
        return underlying().minorVersion();
    }

    @Override
    public List<Field> allFields() {
        return XyzzyxJDIWrapper.wrapFields(virtualMachine(), underlying().allFields());
    }

    @Override
    public List<Location> allLineLocations() throws AbsentInformationException {
        return XyzzyxJDIWrapper.wrapLocations(virtualMachine(), underlying().allLineLocations());
    }

    @Override
    public List<Location> allLineLocations(String stratum, String sourceName) throws AbsentInformationException {
        return XyzzyxJDIWrapper.wrapLocations(virtualMachine(), underlying().allLineLocations(stratum, sourceName));
    }

    @Override
    public List<Method> allMethods() {
        return XyzzyxJDIWrapper.wrapMethods(virtualMachine(), underlying().allMethods());
    }

    @Override
    public List<String> availableStrata() {
        return underlying().availableStrata();
    }

    @Override
    public ClassLoaderReference classLoader() {
        return XyzzyxJDIWrapper.wrap(virtualMachine(), underlying().classLoader());
    }

    @Override
    public ClassObjectReference classObject() {
        return XyzzyxJDIWrapper.wrap(virtualMachine(), underlying().classObject());
    }

    @Override
    public String defaultStratum() {
        return underlying().defaultStratum();
    }

    @Override
    public boolean failedToInitialize() {
        return underlying().failedToInitialize();
    }

    @Override
    public XyzzyxJDIField fieldByName(String name) {
        return XyzzyxJDIWrapper.wrap(virtualMachine(), underlying().fieldByName(name));
    }

    @Override
    public List<Field> fields() {
        return XyzzyxJDIWrapper.wrapFields(virtualMachine(), underlying().fields());
    }

    @Override
    public String genericSignature() {
        return underlying().genericSignature();
    }

    @Override
    public Value getValue(Field field) {
        return XyzzyxJDIWrapper.wrap(virtualMachine(), underlying().getValue(XyzzyxJDIWrapper.unwrap(field)));
    }

    @Override
    public Map<Field, Value> getValues(List<? extends Field> wrappedFields) {
        virtualMachine().setLastFieldAccessException(null);

        // We will find fields which have no getters, and call the underlying
        // getValues to get values for all of them in one fell swoop.
        Map<Field, Field> unwrappedToWrappedMap = new HashMap<Field, Field>();
        List<Field> noGetterUnwrappedFields = new ArrayList<Field>();    // fields that don't have getters

        // But first, for fields that do have getters, call invokeMethod
        // or we will call XyzzyxJDIGetValue for each, depending on doInvokes
        Map<Field, Value> result = new HashMap<Field, Value>();

        // Create the above Maps and lists
        for (Field wrappedField : wrappedFields) {
            Field unwrapped = XyzzyxJDIWrapper.unwrap(wrappedField);
            unwrappedToWrappedMap.put(unwrapped, wrappedField);
            noGetterUnwrappedFields.add(unwrapped);
        }

        // Get values for all the noGetter fields.  Note that this gets them in a single JDWP trip
        Map<Field, Value> unwrappedFieldValues = underlying().getValues(noGetterUnwrappedFields);

        // for each input Field, create a result map entry with that field as the
        // key, and the value returned by getValues, or null if the field is invalid.

        // Make a pass over the unwrapped no getter fields and for each, put its
        // wrapped version, and wrapped value into the result Map.
        for (Map.Entry<Field, Field> unwrappedEntry: unwrappedToWrappedMap.entrySet()) {
            Field wrappedField = unwrappedEntry.getValue();
            Value resultValue = XyzzyxJDIWrapper.wrap(virtualMachine(),
                                             unwrappedFieldValues.get(unwrappedEntry.getKey()));
            result.put(wrappedField, resultValue);
        }
        return result;
    }

    @Override
    public boolean isAbstract() {
        return underlying().isAbstract();
    }

    @Override
    public boolean isFinal() {
        return underlying().isFinal();
    }

    @Override
    public boolean isInitialized() {
        return underlying().isInitialized();
    }

    @Override
    public boolean isPrepared() {
        return underlying().isPrepared();
    }

    @Override
    public boolean isStatic() {
        return underlying().isStatic();
    }

    @Override
    public boolean isVerified() {
        return underlying().isVerified();
    }

    @Override
    public List<Location> locationsOfLine(int lineNumber) throws AbsentInformationException {
        return XyzzyxJDIWrapper.wrapLocations(virtualMachine(), underlying().locationsOfLine(lineNumber));
    }

    @Override
    public List<Location> locationsOfLine(String stratum, String sourceName, int line) throws AbsentInformationException {
        return XyzzyxJDIWrapper.wrapLocations(virtualMachine(), underlying().locationsOfLine(stratum, sourceName, line));
    }

    @Override
    public List<Method> methods() {
        return XyzzyxJDIWrapper.wrapMethods(virtualMachine(), underlying().methods());
    }

    @Override
    public List<Method> methodsByName(String name) {
        return XyzzyxJDIWrapper.wrapMethods(virtualMachine(), underlying().methodsByName(name));
    }

    @Override
    public List<Method> methodsByName(String name, String signature) {
        return XyzzyxJDIWrapper.wrapMethods(virtualMachine(), underlying().methodsByName(name, signature));
    }

    @Override
    public List<ReferenceType> nestedTypes() {
        return XyzzyxJDIWrapper.wrapReferenceTypes(virtualMachine(), underlying().nestedTypes());
    }

    @Override
    public String sourceDebugExtension() throws AbsentInformationException {
        return underlying().sourceDebugExtension();
    }

    @Override
    public String sourceName() throws AbsentInformationException {
        return underlying().sourceName();
    }

    @Override
    public List<String> sourceNames(String stratum) throws AbsentInformationException {
        return underlying().sourceNames(stratum);
    }

    @Override
    public List<String> sourcePaths(String stratum) throws AbsentInformationException {
        return underlying().sourcePaths(stratum);
    }

    @Override
    public List<Field> visibleFields() {
        return XyzzyxJDIWrapper.wrapFields(virtualMachine(), underlying().visibleFields());
    }

    @Override
    public List<Method> visibleMethods() {
        return XyzzyxJDIWrapper.wrapMethods(virtualMachine(), underlying().visibleMethods());
    }

    @Override
    public int compareTo(ReferenceType o) {
        return underlying().compareTo(XyzzyxJDIWrapper.unwrap(o));
    }

    @Override
    public boolean isPackagePrivate() {
        return underlying().isPackagePrivate();
    }

    @Override
    public boolean isPrivate() {
        return underlying().isPrivate();
    }

    @Override
    public boolean isProtected() {
        return underlying().isProtected();
    }

    @Override
    public boolean isPublic() {
        return underlying().isPublic();
    }

    @Override
    public int modifiers() {
        return underlying().modifiers();
    }

    @Override
    protected ReferenceType underlying() {
        return (ReferenceType) super.underlying();
    }

    public ReferenceType _underlying() {
        return (ReferenceType)super.underlying();
    }
}
