/*
 * Copyright (c) 2010-2012, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.oracle.xyzzyx.jdi;

import com.sun.jdi.AbsentInformationException;
import com.sun.jdi.ClassNotLoadedException;
import com.sun.jdi.IncompatibleThreadStateException;
import com.sun.jdi.InvalidStackFrameException;
import com.sun.jdi.InvalidTypeException;
import com.sun.jdi.MonitorInfo;
import com.sun.jdi.ObjectReference;
import com.sun.jdi.StackFrame;
import com.sun.jdi.ThreadReference;
import com.sun.jdi.Value;

import java.util.ArrayList;
import java.util.List;

/**
 *
 */
public class XyzzyxJDIThreadReference extends XyzzyxJDIObjectReference implements ThreadReference {
    public XyzzyxJDIThreadReference(XyzzyxJDIVirtualMachine xyzzyxvm, ThreadReference underlying) {
        super(xyzzyxvm, underlying);
    }

    @Override
    public void forceEarlyReturn(Value value) throws InvalidTypeException, ClassNotLoadedException, IncompatibleThreadStateException {
        underlying().forceEarlyReturn(XyzzyxJDIWrapper.unwrap(value));
    }

    @Override
    public List<MonitorInfo> ownedMonitorsAndFrames() throws IncompatibleThreadStateException {
        return XyzzyxJDIWrapper.wrapMonitorInfos(virtualMachine(), underlying().ownedMonitorsAndFrames());
    }

    @Override
    public XyzzyxJDIObjectReference currentContendedMonitor() throws IncompatibleThreadStateException {
        return XyzzyxJDIWrapper.wrap(virtualMachine(), underlying().currentContendedMonitor());
    }

    @Override
    public XyzzyxJDIStackFrame frame(int index) throws IncompatibleThreadStateException {
        List<StackFrame> frames = getFilteredFrames();
        if (index < 0 || index >= frames.size()) {
            throw new IndexOutOfBoundsException();
        }
        return (XyzzyxJDIStackFrame) frames.get(index);
    }

    @Override
    public int frameCount() throws IncompatibleThreadStateException {
        return getFilteredFrames().size();
    }

    @Override
    public List<StackFrame> frames() throws IncompatibleThreadStateException {
        return getFilteredFrames();
    }

    @Override
    public List<StackFrame> frames(int start, int length) throws IncompatibleThreadStateException {
        List<StackFrame> frames = getFilteredFrames();
        int frameCount = frames.size();
        if (start < 0 || start >= frameCount || length < 0 || (start+length) > frameCount) {
            throw new IndexOutOfBoundsException();
        }
        List<StackFrame> result = new ArrayList<>();
        for (int count = start; count < (start+length); count++) {
            result.add(frames.get(count));
        }
        return result;
    }

    @Override
    public void interrupt() {
        underlying().interrupt();
    }

    @Override
    public boolean isAtBreakpoint() {
        return underlying().isAtBreakpoint();
    }

    @Override
    public boolean isSuspended() {
        return underlying().isSuspended();
    }

    @Override
    public String name() {
        return underlying().name();
    }

    @Override
    public List<ObjectReference> ownedMonitors() throws IncompatibleThreadStateException {
        return XyzzyxJDIWrapper.wrapObjectReferences(virtualMachine(), underlying().ownedMonitors());
    }

    @Override
    public void popFrames(StackFrame frame) throws IncompatibleThreadStateException {
        underlying().popFrames(XyzzyxJDIWrapper.unwrap(frame));
    }

    @Override
    public void resume() {
        underlying().resume();
    }

    @Override
    public int status() {
        return underlying().status();
    }

    @Override
    public void stop(ObjectReference exception) throws InvalidTypeException {
        underlying().stop(XyzzyxJDIWrapper.unwrap(exception));
    }

    @Override
    public void suspend() {
        underlying().suspend();
    }

    @Override
    public int suspendCount() {
        return underlying().suspendCount();
    }

    @Override
    public XyzzyxJDIThreadGroupReference threadGroup() {
        return XyzzyxJDIWrapper.wrap(virtualMachine(), underlying().threadGroup());
    }

    @Override
    protected ThreadReference underlying() {
        return (ThreadReference) super.underlying();
    }

    private List<StackFrame> getFilteredFrames() throws IncompatibleThreadStateException {
        List<StackFrame> frames = XyzzyxJDIWrapper.wrapFrames(virtualMachine(), underlying().frames());
        List<StackFrame> filteredFrames = new ArrayList<>(frames.size());
        try {
            for (StackFrame fr : frames) {
                XyzzyxJDIStackFrame fxfr = (XyzzyxJDIStackFrame) fr;
                // Filter frames here.
                filteredFrames.add(fxfr);
            }
        } catch (InvalidStackFrameException exp) {
            throw new IncompatibleThreadStateException(exp.getMessage());
        }
        return filteredFrames;
    }
}
