/*
 * Copyright (c) 2010-2012, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.oracle.xyzzyx.jdi;

import com.oracle.xyzzyx.jdi.event.XyzzyxJDIEventQueue;
import com.oracle.xyzzyx.jdi.request.XyzzyxJDIEventRequestManager;

import com.sun.jdi.ArrayReference;
import com.sun.jdi.ArrayType;
import com.sun.jdi.BooleanType;
import com.sun.jdi.BooleanValue;
import com.sun.jdi.ByteType;
import com.sun.jdi.ByteValue;
import com.sun.jdi.CharType;
import com.sun.jdi.CharValue;
import com.sun.jdi.ClassLoaderReference;
import com.sun.jdi.ClassObjectReference;
import com.sun.jdi.ClassType;
import com.sun.jdi.DoubleType;
import com.sun.jdi.DoubleValue;
import com.sun.jdi.Field;
import com.sun.jdi.FloatType;
import com.sun.jdi.FloatValue;
import com.sun.jdi.IntegerType;
import com.sun.jdi.IntegerValue;
import com.sun.jdi.InterfaceType;
import com.sun.jdi.LocalVariable;
import com.sun.jdi.Location;
import com.sun.jdi.LongType;
import com.sun.jdi.LongValue;
import com.sun.jdi.Method;
import com.sun.jdi.MonitorInfo;
import com.sun.jdi.ObjectReference;
import com.sun.jdi.PrimitiveType;
import com.sun.jdi.PrimitiveValue;
import com.sun.jdi.ReferenceType;
import com.sun.jdi.ShortType;
import com.sun.jdi.ShortValue;
import com.sun.jdi.StackFrame;
import com.sun.jdi.StringReference;
import com.sun.jdi.ThreadGroupReference;
import com.sun.jdi.ThreadReference;
import com.sun.jdi.Type;
import com.sun.jdi.Value;
import com.sun.jdi.VirtualMachine;
import com.sun.jdi.VoidType;
import com.sun.jdi.VoidValue;
import com.sun.jdi.event.EventQueue;
import com.sun.jdi.request.EventRequestManager;

import java.util.ArrayList;
import java.util.List;

/**
 *
 */
public class XyzzyxJDIWrapper {
    public static XyzzyxJDIVirtualMachine wrap(VirtualMachine vm) {
        return (vm == null)? null : new XyzzyxJDIVirtualMachine(vm);
    }

    public static List<VirtualMachine> wrapVirtualMachines(List<VirtualMachine> vms) {
        List<VirtualMachine> res = new ArrayList<>(vms.size());
        for (VirtualMachine vm : vms) {
            res.add(wrap(vm));
        }
        return res;
    }

    public static XyzzyxJDIType wrap(XyzzyxJDIVirtualMachine xyzzyxvm, Type type) {
        if (type == null) {
            return null;
        }

        if (type instanceof VoidType) {
            return xyzzyxvm.voidType((VoidType)type);
        } else if (type instanceof PrimitiveType) {
            if (type instanceof BooleanType) {
                return xyzzyxvm.booleanType((BooleanType)type);
            } else if (type instanceof CharType) {
                return xyzzyxvm.charType((CharType)type);
            } else if (type instanceof ByteType) {
                return xyzzyxvm.byteType((ByteType)type);
            } else if (type instanceof ShortType) {
                return xyzzyxvm.shortType((ShortType)type);
            } else if (type instanceof IntegerType) {
                return xyzzyxvm.integerType((IntegerType)type);
            } else if (type instanceof LongType) {
                return xyzzyxvm.longType((LongType)type);
            } else if (type instanceof FloatType) {
                return xyzzyxvm.floatType((FloatType)type);
            } else if (type instanceof DoubleType) {
                return xyzzyxvm.doubleType((DoubleType)type);
            } else {
                throw new IllegalArgumentException("illegal primitive type : " + type);
            }
        } else if (type instanceof ReferenceType) {
            return wrap(xyzzyxvm, (ReferenceType)type);
        } else {
            throw new IllegalArgumentException("illegal type: " + type);
        }
    }

    public static List<Type> wrapTypes(XyzzyxJDIVirtualMachine xyzzyxvm, List<Type> types) {
        if (types == null) {
            return null;
        }
        List<Type> result = new ArrayList<>(types.size());
        for (Type type : types) {
            result.add(wrap(xyzzyxvm, type));
        }
        return result;
    }

    public static XyzzyxJDIReferenceType wrap(XyzzyxJDIVirtualMachine xyzzyxvm, ReferenceType rt) {
        if (rt == null) {
            return null;
        } else if (rt instanceof ClassType) {
            return xyzzyxvm.classType((ClassType)rt);
        } else if (rt instanceof InterfaceType) {
            return xyzzyxvm.interfaceType((InterfaceType)rt);
        } else if (rt instanceof ArrayType) {
            return xyzzyxvm.arrayType((ArrayType)rt);
        } else {
            return xyzzyxvm.referenceType(rt);
        }
    }

    public static XyzzyxJDIClassType wrap(XyzzyxJDIVirtualMachine xyzzyxvm, ClassType ct) {
        return (ct == null)? null : xyzzyxvm.classType(ct);
    }

    public static XyzzyxJDIInterfaceType wrap(XyzzyxJDIVirtualMachine xyzzyxvm, InterfaceType it) {
        return (it == null)? null : xyzzyxvm.interfaceType(it);
    }

    public static XyzzyxJDIArrayType wrap(XyzzyxJDIVirtualMachine xyzzyxvm, ArrayType at) {
        return (at == null)? null : xyzzyxvm.arrayType(at);
    }

    public static List<ReferenceType> wrapReferenceTypes(XyzzyxJDIVirtualMachine xyzzyxvm, List<ReferenceType> refTypes) {
        // Note that VirtualMachineImpl caches the list, and returns an unmodifiable wrapped list.
        // Classes that get loaded in the future are added to its list by an EventListener on ClassPrepared
        // events.  If we cache our wrapped list, they we would have to do the same thing, or be able
        // to update our cached list when this method is called again.  So for the time being,
        // we won't cache and thus don't have to return an unmodifiable list.
        if (refTypes == null) {
            return null;
        }
        List<ReferenceType> result = new ArrayList<>(refTypes.size());
        for (ReferenceType rt : refTypes) {
            String className = rt.name();
            result.add(XyzzyxJDIWrapper.wrap(xyzzyxvm, rt));
        }
        return result;
    }

    public static List<ClassType> wrapClassTypes(XyzzyxJDIVirtualMachine xyzzyxvm, List<ClassType> classes) {
        if (classes == null) {
            return null;
        }
        List<ClassType> result = new ArrayList<>(classes.size());
        for (ClassType ct : classes) {
            result.add(XyzzyxJDIWrapper.wrap(xyzzyxvm, ct));
        }
        return result;
    }

    public static List<InterfaceType> wrapInterfaceTypes(XyzzyxJDIVirtualMachine xyzzyxvm, List<InterfaceType> interfaces) {
        if (interfaces == null) {
            return null;
        }
        List<InterfaceType> result = new ArrayList<>(interfaces.size());
        for (InterfaceType it : interfaces) {
            result.add(XyzzyxJDIWrapper.wrap(xyzzyxvm, it));
        }
        return result;
    }

    public static XyzzyxJDILocation wrap(XyzzyxJDIVirtualMachine xyzzyxvm, Location loc) {
        return (loc == null)? null : xyzzyxvm.location(loc);
    }

    public static List<Location> wrapLocations(XyzzyxJDIVirtualMachine xyzzyxvm, List<Location> locations) {
        if (locations == null) {
            return null;
        }
        List<Location> result = new ArrayList<>(locations.size());
        for (Location loc: locations) {
            result.add(wrap(xyzzyxvm, loc));
        }
        return result;
    }

    public static XyzzyxJDIField wrap(XyzzyxJDIVirtualMachine xyzzyxvm, Field field) {
        return (field == null)? null : xyzzyxvm.field(field);
    }

    /*
     * The fields are JDI Fields.
     * Each field can be a user field of an XyzzyxJDI class, an internal field of an XyzzyxJDI class,
     * or a field of a Java class.
     */
    public static List<Field> wrapFields(XyzzyxJDIVirtualMachine xyzzyxvm, List<Field> fields) {
        // Create XyzzyxJDIField wrappers for each field that is a valid XyzzyxJDI field.
        if (fields == null) {
            return null;
        }
        // We will have far fewer fields than fields.size() due to all the VFLGS etc
        // fields we will discard , so start with some small random amount
        List<Field> result = new ArrayList<>(20);

        for (Field fld : fields) {
            result.add(xyzzyxvm.field(fld));
        }
        return result;
    }

    public static XyzzyxJDIMethod wrap(XyzzyxJDIVirtualMachine xyzzyxvm, Method method) {
        return (method == null)? null : xyzzyxvm.method(method);
    }

    public static List<Method> wrapMethods(XyzzyxJDIVirtualMachine xyzzyxvm, List<Method> methods) {
        if (methods == null) {
            return null;
        }
        List<Method> result = new ArrayList<>(20);
        for (Method mth : methods) {
            result.add(xyzzyxvm.method(mth));
        }
        return result;
    }

    public static XyzzyxJDIMonitorInfo wrap(XyzzyxJDIVirtualMachine xyzzyxvm, MonitorInfo monitorInfo) {
        return (monitorInfo == null)? null : xyzzyxvm.monitorInfo(monitorInfo);
    }

    public static List<MonitorInfo> wrapMonitorInfos(XyzzyxJDIVirtualMachine xyzzyxvm, List<MonitorInfo> monInfos) {
        if (monInfos == null) {
            return null;
        }
        List<MonitorInfo> result = new ArrayList<>(monInfos.size());
        for (MonitorInfo mi : monInfos) {
            result.add(wrap(xyzzyxvm, mi));
        }
        return result;
    }

    public static XyzzyxJDIStackFrame wrap(XyzzyxJDIVirtualMachine xyzzyxvm, StackFrame frame) {
        return (frame == null)? null : xyzzyxvm.stackFrame(frame);
    }

    public static List<StackFrame> wrapFrames(XyzzyxJDIVirtualMachine xyzzyxvm, List<StackFrame> frames) {
        if (frames == null) {
            return null;
        }
        List<StackFrame> result = new ArrayList<>(frames.size());
        for (StackFrame fr : frames) {
            result.add(wrap(xyzzyxvm, fr));
        }
        return result;
    }

    public static XyzzyxJDILocalVariable wrap(XyzzyxJDIVirtualMachine xyzzyxvm, LocalVariable var) {
        return (var == null)? null : xyzzyxvm.localVariable(var);
    }

    public static List<LocalVariable> wrapLocalVariables(XyzzyxJDIVirtualMachine xyzzyxvm, List<LocalVariable> locals) {
        if (locals == null) {
            return null;
        }
        List<LocalVariable> result = new ArrayList<>(locals.size());
        for (LocalVariable var: locals) {
            result.add(wrap(xyzzyxvm, var));
        }
        return result;
    }

    public static XyzzyxJDIValue wrap(XyzzyxJDIVirtualMachine xyzzyxvm, Value value) {
        if (value == null) {
            return null;
        }

        if (value instanceof PrimitiveValue) {
            if (value instanceof BooleanValue) {
                return xyzzyxvm.booleanValue((BooleanValue)value);
            } else if (value instanceof CharValue) {
                return xyzzyxvm.charValue((CharValue)value);
            } else if (value instanceof ByteValue) {
                return xyzzyxvm.byteValue((ByteValue)value);
            } else if (value instanceof ShortValue) {
                return xyzzyxvm.shortValue((ShortValue)value);
            } else if (value instanceof IntegerValue) {
                return xyzzyxvm.integerValue((IntegerValue)value);
            } else if (value instanceof LongValue) {
                return xyzzyxvm.longValue((LongValue)value);
            } else if (value instanceof FloatValue) {
                return xyzzyxvm.floatValue((FloatValue)value);
            } else if (value instanceof DoubleValue) {
                return xyzzyxvm.doubleValue((DoubleValue)value);
            } else {
                throw new IllegalArgumentException("illegal primitive value : " + value);
            }
        } else if (value instanceof VoidValue) {
            return xyzzyxvm.voidValue();
        } else if (value instanceof ObjectReference) {
            return  wrap(xyzzyxvm, (ObjectReference)value);
        } else {
            throw new IllegalArgumentException("illegal value: " + value);
        }
    }

    public static List<ObjectReference> wrapObjectReferences(XyzzyxJDIVirtualMachine xyzzyxvm, List<ObjectReference> refs) {
        if (refs == null) {
            return null;
        }
        List<ObjectReference> result = new ArrayList<>(refs.size());
        for (ObjectReference ref : refs) {
            result.add(wrap(xyzzyxvm, ref));
        }
        return result;
    }


    public static XyzzyxJDIObjectReference wrap(XyzzyxJDIVirtualMachine xyzzyxvm, ObjectReference ref) {
        if (ref == null) {
            return null;
        } else if (ref instanceof ArrayReference) {
            return xyzzyxvm.arrayReference((ArrayReference)ref);
        } else if (ref instanceof StringReference) {
            return xyzzyxvm.stringReference((StringReference)ref);
        } else if (ref instanceof ThreadReference) {
            return xyzzyxvm.threadReference((ThreadReference)ref);
        } else if (ref instanceof ThreadGroupReference) {
            return xyzzyxvm.threadGroupReference((ThreadGroupReference)ref);
        } else if (ref instanceof ClassLoaderReference) {
            return xyzzyxvm.classLoaderReference((ClassLoaderReference)ref);
        } else if (ref instanceof ClassObjectReference) {
            return xyzzyxvm.classObjectReference((ClassObjectReference)ref);
        } else {
            // Massage objects here.
            return xyzzyxvm.objectReference(ref);
        }
    }

    public static XyzzyxJDIArrayReference wrap(XyzzyxJDIVirtualMachine xyzzyxvm, ArrayReference ref) {
        return (ref == null)? null : xyzzyxvm.arrayReference(ref);
    }

    public static XyzzyxJDIThreadReference wrap(XyzzyxJDIVirtualMachine xyzzyxvm, ThreadReference ref) {
        return (ref == null)? null : xyzzyxvm.threadReference(ref);
    }


    public static XyzzyxJDIThreadGroupReference wrap(XyzzyxJDIVirtualMachine xyzzyxvm, ThreadGroupReference ref) {
        return (ref == null)? null : xyzzyxvm.threadGroupReference(ref);
    }

    public static List<ThreadReference> wrapThreads(XyzzyxJDIVirtualMachine xyzzyxvm, List<ThreadReference> threads) {
        if (threads == null) {
            return null;
        }
        List<ThreadReference> result = new ArrayList<>(threads.size());
        for (ThreadReference tref : threads) {
            result.add(wrap(xyzzyxvm, tref));
        }
        return result;
    }

    public static List<ThreadGroupReference> wrapThreadGroups(XyzzyxJDIVirtualMachine xyzzyxvm, List<ThreadGroupReference> threadGroups) {
        if (threadGroups == null) {
            return null;
        }
        List<ThreadGroupReference> result = new ArrayList<>(threadGroups.size());
        for (ThreadGroupReference tref : threadGroups) {
            result.add(wrap(xyzzyxvm, tref));
        }
        return result;
    }

    public static XyzzyxJDIClassLoaderReference wrap(XyzzyxJDIVirtualMachine xyzzyxvm, ClassLoaderReference ref) {
        return (ref == null)? null : xyzzyxvm.classLoaderReference(ref);
    }

    public static XyzzyxJDIClassObjectReference wrap(XyzzyxJDIVirtualMachine xyzzyxvm, ClassObjectReference ref) {
        return (ref == null)? null : xyzzyxvm.classObjectReference(ref);
    }

    public static List<Value> wrapValues(XyzzyxJDIVirtualMachine xyzzyxvm, List<Value> values) {
        if (values == null) {
            return null;
        }
        List<Value> result = new ArrayList<>(values.size());
        for (Value v : values) {
            result.add(wrap(xyzzyxvm, v));
        }
        return result;
    }

    public static Location unwrap(Location loc) {
        return (loc instanceof XyzzyxJDILocation)? ((XyzzyxJDILocation)loc).underlying() : loc;
    }

    public static StackFrame unwrap(StackFrame frame) {
        return (frame instanceof XyzzyxJDIStackFrame)? ((XyzzyxJDIStackFrame)frame).underlying() : frame;
    }

    public static LocalVariable unwrap(LocalVariable var) {
        return (var instanceof XyzzyxJDILocalVariable)? ((XyzzyxJDILocalVariable)var).underlying() : var;
    }

    public static Value unwrap(Value value) {
        return (value instanceof XyzzyxJDIValue)? ((XyzzyxJDIValue)value).underlying() : value;
    }

    public static List<? extends Value> unwrapValues(List<? extends Value> values) {
        if (values == null) {
            return null;
        }
        List<Value> result = new ArrayList<>(values.size());
        for (Value v : values) {
            result.add(unwrap(v));
        }
        return result;
    }

    public static Field unwrap(Field field) {
        return (field instanceof XyzzyxJDIField)? ((XyzzyxJDIField)field).underlying() : field;
    }

    public static Method unwrap(Method method) {
        return (method instanceof XyzzyxJDIMethod)? ((XyzzyxJDIMethod)method).underlying() : method;
    }

    public static ObjectReference unwrap(ObjectReference ref) {
        return (ref instanceof XyzzyxJDIObjectReference)? ((XyzzyxJDIObjectReference)ref).underlying() : ref;
    }

    public static ThreadReference unwrap(ThreadReference ref) {
        return (ref instanceof XyzzyxJDIThreadReference)? ((XyzzyxJDIThreadReference)ref).underlying() : ref;
    }

    public static ReferenceType unwrap(ReferenceType rt) {
        return (rt instanceof XyzzyxJDIReferenceType)? ((XyzzyxJDIReferenceType)rt).underlying() : rt;
    }

    public static List<? extends ReferenceType> unwrapReferenceTypes(List<? extends ReferenceType> refTypes) {
        if (refTypes == null) {
            return null;
        }
        List<ReferenceType> result = new ArrayList<ReferenceType>(refTypes.size());
        for (ReferenceType rt : refTypes) {
            result.add(unwrap(rt));
        }
        return result;
    }

    // event requests
    public static XyzzyxJDIEventRequestManager wrap(XyzzyxJDIVirtualMachine xyzzyxvm, EventRequestManager man) {
        return (man == null)? null : new XyzzyxJDIEventRequestManager(xyzzyxvm, man);
    }

    // event queue
    public static XyzzyxJDIEventQueue wrap(XyzzyxJDIVirtualMachine xyzzyxvm, EventQueue evtQueue) {
        return (evtQueue == null)? null : new XyzzyxJDIEventQueue(xyzzyxvm, evtQueue);
    }
}
