/*
 * Copyright (c) 2010-2012, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.oracle.xyzzyx.jdi.event;

import com.oracle.xyzzyx.jdi.XyzzyxJDIMirror;
import com.oracle.xyzzyx.jdi.XyzzyxJDIVirtualMachine;
import com.oracle.xyzzyx.jdi.request.XyzzyxJDIEventRequest;

import com.sun.jdi.event.AccessWatchpointEvent;
import com.sun.jdi.event.BreakpointEvent;
import com.sun.jdi.event.ClassPrepareEvent;
import com.sun.jdi.event.ClassUnloadEvent;
import com.sun.jdi.event.Event;
import com.sun.jdi.event.ExceptionEvent;
import com.sun.jdi.event.LocatableEvent;
import com.sun.jdi.event.MethodEntryEvent;
import com.sun.jdi.event.MethodExitEvent;
import com.sun.jdi.event.ModificationWatchpointEvent;
import com.sun.jdi.event.MonitorContendedEnterEvent;
import com.sun.jdi.event.MonitorContendedEnteredEvent;
import com.sun.jdi.event.MonitorWaitEvent;
import com.sun.jdi.event.MonitorWaitedEvent;
import com.sun.jdi.event.StepEvent;
import com.sun.jdi.event.ThreadDeathEvent;
import com.sun.jdi.event.ThreadStartEvent;
import com.sun.jdi.event.VMDeathEvent;
import com.sun.jdi.event.VMDisconnectEvent;
import com.sun.jdi.event.VMStartEvent;
import com.sun.jdi.event.WatchpointEvent;

import java.util.ArrayList;
import java.util.Collection;

/**
 *
 */
public class XyzzyxJDIEvent extends XyzzyxJDIMirror implements Event {
    public XyzzyxJDIEvent(XyzzyxJDIVirtualMachine xyzzyxvm, Event underlying) {
        super(xyzzyxvm, underlying);
    }

    public XyzzyxJDIEventRequest request() {
        return XyzzyxJDIEventRequest.wrap(virtualMachine(), underlying().request());
    }

    @Override
    protected Event underlying() {
        return (Event) super.underlying();
    }

    public static Event unwrap(Object evt) {
        return (evt instanceof XyzzyxJDIEvent)? ((XyzzyxJDIEvent)evt).underlying() : (Event)evt;
    }

    public static XyzzyxJDIEvent wrap(XyzzyxJDIVirtualMachine xyzzyxvm, Event evt) {
        if (evt == null) {
            return null;
        }
        if (evt instanceof AccessWatchpointEvent) {
            return new XyzzyxJDIAccessWatchpointEvent(xyzzyxvm, (AccessWatchpointEvent)evt);
        } else if (evt instanceof BreakpointEvent) {
            return new XyzzyxJDIBreakpointEvent(xyzzyxvm, (BreakpointEvent)evt);
        } else if (evt instanceof ClassPrepareEvent) {
            return new XyzzyxJDIClassPrepareEvent(xyzzyxvm, (ClassPrepareEvent)evt);
        } else if (evt instanceof ClassUnloadEvent) {
            return new XyzzyxJDIClassUnloadEvent(xyzzyxvm, (ClassUnloadEvent)evt);
        } else if (evt instanceof ExceptionEvent) {
            return new XyzzyxJDIExceptionEvent(xyzzyxvm, (ExceptionEvent)evt);
        } else if (evt instanceof MethodEntryEvent) {
            return new XyzzyxJDIMethodEntryEvent(xyzzyxvm, (MethodEntryEvent)evt);
        } else if (evt instanceof MethodExitEvent) {
            return new XyzzyxJDIMethodExitEvent(xyzzyxvm, (MethodExitEvent)evt);
        } else if (evt instanceof ModificationWatchpointEvent) {
            return new XyzzyxJDIModificationWatchpointEvent(xyzzyxvm, (ModificationWatchpointEvent)evt);
        } else if (evt instanceof MonitorContendedEnterEvent) {
            return new XyzzyxJDIMonitorContendedEnterEvent(xyzzyxvm, (MonitorContendedEnterEvent)evt);
        } else if (evt instanceof MonitorContendedEnteredEvent) {
            return new XyzzyxJDIMonitorContendedEnteredEvent(xyzzyxvm, (MonitorContendedEnteredEvent)evt);
        } else if (evt instanceof MonitorWaitEvent) {
            return new XyzzyxJDIMonitorWaitEvent(xyzzyxvm, (MonitorWaitEvent)evt);
        } else if (evt instanceof MonitorWaitedEvent) {
            return new XyzzyxJDIMonitorWaitedEvent(xyzzyxvm, (MonitorWaitedEvent)evt);
        } else if (evt instanceof StepEvent) {
            return new XyzzyxJDIStepEvent(xyzzyxvm, (StepEvent)evt);
        } else if (evt instanceof ThreadDeathEvent) {
            return new XyzzyxJDIThreadDeathEvent(xyzzyxvm, (ThreadDeathEvent)evt);
        } else if (evt instanceof ThreadStartEvent) {
            return new XyzzyxJDIThreadStartEvent(xyzzyxvm, (ThreadStartEvent)evt);
        } else if (evt instanceof VMDeathEvent) {
            return new XyzzyxJDIVMDeathEvent(xyzzyxvm, (VMDeathEvent)evt);
        } else if (evt instanceof VMDisconnectEvent) {
            return new XyzzyxJDIVMDisconnectEvent(xyzzyxvm, (VMDisconnectEvent)evt);
        } else if (evt instanceof VMStartEvent) {
            return new XyzzyxJDIVMStartEvent(xyzzyxvm, (VMStartEvent)evt);
        } else if (evt instanceof WatchpointEvent) {
            return new XyzzyxJDIWatchpointEvent(xyzzyxvm, (WatchpointEvent)evt);
        } else if (evt instanceof LocatableEvent) {
            return new XyzzyxJDILocatableEvent(xyzzyxvm, (LocatableEvent)evt);
        } else {
            return new XyzzyxJDIEvent(xyzzyxvm, evt);
        }
    }

    public static Collection<? extends Event> unwrapEvents(Collection<?> events) {
        if (events == null) {
            return null;
        }
        ArrayList<Event> result = new ArrayList<>();
        for (Object obj : events) {
            result.add(((Event)((obj instanceof Event) ? unwrap(obj) : obj)));
        }
        return result;
    }
}