/*
 * Copyright (c) 2010-2012, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package com.oracle.xyzzyx.jdi.request;

import com.oracle.xyzzyx.jdi.XyzzyxJDIMirror;
import com.oracle.xyzzyx.jdi.XyzzyxJDIVirtualMachine;

import com.sun.jdi.request.AccessWatchpointRequest;
import com.sun.jdi.request.BreakpointRequest;
import com.sun.jdi.request.ClassPrepareRequest;
import com.sun.jdi.request.ClassUnloadRequest;
import com.sun.jdi.request.EventRequest;
import com.sun.jdi.request.ExceptionRequest;
import com.sun.jdi.request.MethodEntryRequest;
import com.sun.jdi.request.MethodExitRequest;
import com.sun.jdi.request.ModificationWatchpointRequest;
import com.sun.jdi.request.MonitorContendedEnterRequest;
import com.sun.jdi.request.MonitorContendedEnteredRequest;
import com.sun.jdi.request.MonitorWaitRequest;
import com.sun.jdi.request.MonitorWaitedRequest;
import com.sun.jdi.request.StepRequest;
import com.sun.jdi.request.ThreadDeathRequest;
import com.sun.jdi.request.ThreadStartRequest;
import com.sun.jdi.request.VMDeathRequest;
import com.sun.jdi.request.WatchpointRequest;

import java.util.ArrayList;
import java.util.List;

/**
 *
 */
public class XyzzyxJDIEventRequest extends XyzzyxJDIMirror implements EventRequest {
    public XyzzyxJDIEventRequest(XyzzyxJDIVirtualMachine xyzzyxvm, EventRequest underlying) {
        super(xyzzyxvm, underlying);
    }

    @Override
    public void addCountFilter(int count) {
        underlying().addCountFilter(count);
    }

    @Override
    public void disable() {
        underlying().disable();
    }

    @Override
    public void enable() {
        underlying().enable();
    }

    @Override
    public Object getProperty(Object arg0) {
        return underlying().getProperty(arg0);
    }

    @Override
    public boolean isEnabled() {
        return underlying().isEnabled();
    }

    @Override
    public void putProperty(Object arg0, Object arg1) {
        underlying().putProperty(arg0, arg1);
    }

    @Override
    public void setEnabled(boolean enabled) {
        underlying().setEnabled(enabled);
    }

    @Override
    public void setSuspendPolicy(int policy) {
        underlying().setSuspendPolicy(policy);
    }

    @Override
    public int suspendPolicy() {
        return underlying().suspendPolicy();
    }

    @Override
    protected EventRequest underlying() {
        return (EventRequest) super.underlying();
    }

    // static utils for wrapping/unwrapping event request objects

    public static XyzzyxJDIEventRequest wrap(XyzzyxJDIVirtualMachine xyzzyxvm, EventRequest req) {
        if (req == null) {
            return null;
        }

        if (req instanceof AccessWatchpointRequest) {
            return new XyzzyxJDIAccessWatchpointRequest(xyzzyxvm, (AccessWatchpointRequest)req);
        } else if (req instanceof BreakpointRequest) {
            return new XyzzyxJDIBreakpointRequest(xyzzyxvm, (BreakpointRequest)req);
        } else if (req instanceof ClassPrepareRequest) {
            return new XyzzyxJDIClassPrepareRequest(xyzzyxvm, (ClassPrepareRequest)req);
        } else if (req instanceof ClassUnloadRequest) {
            return new XyzzyxJDIClassUnloadRequest(xyzzyxvm, (ClassUnloadRequest)req);
        } else if (req instanceof ExceptionRequest) {
            return new XyzzyxJDIExceptionRequest(xyzzyxvm, (ExceptionRequest)req);
        } else if (req instanceof MethodEntryRequest) {
            return new XyzzyxJDIMethodEntryRequest(xyzzyxvm, (MethodEntryRequest)req);
        } else if (req instanceof MethodExitRequest) {
            return new XyzzyxJDIMethodExitRequest(xyzzyxvm, (MethodExitRequest)req);
        } else if (req instanceof ModificationWatchpointRequest) {
            return new XyzzyxJDIModificationWatchpointRequest(xyzzyxvm, (ModificationWatchpointRequest)req);
        } else if (req instanceof MonitorContendedEnterRequest) {
            return new XyzzyxJDIMonitorContendedEnterRequest(xyzzyxvm, (MonitorContendedEnterRequest)req);
        } else if (req instanceof MonitorContendedEnteredRequest) {
            return new XyzzyxJDIMonitorContendedEnteredRequest(xyzzyxvm, (MonitorContendedEnteredRequest)req);
        } else if (req instanceof MonitorWaitRequest) {
            return new XyzzyxJDIMonitorWaitRequest(xyzzyxvm, (MonitorWaitRequest)req);
        } else if (req instanceof MonitorWaitedRequest) {
            return new XyzzyxJDIMonitorWaitedRequest(xyzzyxvm, (MonitorWaitedRequest)req);
        } else if (req instanceof StepRequest) {
            return new XyzzyxJDIStepRequest(xyzzyxvm, (StepRequest)req);
        } else if (req instanceof ThreadDeathRequest) {
            return new XyzzyxJDIThreadDeathRequest(xyzzyxvm, (ThreadDeathRequest)req);
        } else if (req instanceof ThreadStartRequest) {
            return new XyzzyxJDIThreadStartRequest(xyzzyxvm, (ThreadStartRequest)req);
        } else if (req instanceof VMDeathRequest) {
            return new XyzzyxJDIVMDeathRequest(xyzzyxvm, (VMDeathRequest)req);
        } else if (req instanceof WatchpointRequest) {
            return new XyzzyxJDIWatchpointRequest(xyzzyxvm, (WatchpointRequest)req);
        } else {
            return new XyzzyxJDIEventRequest(xyzzyxvm, req);
        }
    }

    public static XyzzyxJDIAccessWatchpointRequest wrap(
            XyzzyxJDIVirtualMachine xyzzyxvm, AccessWatchpointRequest req) {
        return (req == null)? null : new XyzzyxJDIAccessWatchpointRequest(xyzzyxvm, req);
    }

    public static XyzzyxJDIBreakpointRequest wrap(
            XyzzyxJDIVirtualMachine xyzzyxvm, BreakpointRequest req) {
        return (req == null)? null : new XyzzyxJDIBreakpointRequest(xyzzyxvm, req);
    }

    public static XyzzyxJDIClassPrepareRequest wrap(
            XyzzyxJDIVirtualMachine xyzzyxvm, ClassPrepareRequest req) {
        return (req == null)? null : new XyzzyxJDIClassPrepareRequest(xyzzyxvm, req);
    }

    public static XyzzyxJDIClassUnloadRequest wrap(
            XyzzyxJDIVirtualMachine xyzzyxvm, ClassUnloadRequest req) {
        return (req == null)? null : new XyzzyxJDIClassUnloadRequest(xyzzyxvm, req);
    }

    public static XyzzyxJDIExceptionRequest wrap(
            XyzzyxJDIVirtualMachine xyzzyxvm, ExceptionRequest req) {
        return (req == null)? null : new XyzzyxJDIExceptionRequest(xyzzyxvm, req);
    }

    public static XyzzyxJDIMethodEntryRequest wrap(
            XyzzyxJDIVirtualMachine xyzzyxvm, MethodEntryRequest req) {
        return (req == null)? null : new XyzzyxJDIMethodEntryRequest(xyzzyxvm, req);
    }

    public static XyzzyxJDIMethodExitRequest wrap(
            XyzzyxJDIVirtualMachine xyzzyxvm, MethodExitRequest req) {
        return (req == null)? null : new XyzzyxJDIMethodExitRequest(xyzzyxvm, req);
    }

    public static XyzzyxJDIModificationWatchpointRequest wrap(
            XyzzyxJDIVirtualMachine xyzzyxvm, ModificationWatchpointRequest req) {
        return (req == null)? null : new XyzzyxJDIModificationWatchpointRequest(xyzzyxvm, req);
    }

    public static XyzzyxJDIMonitorContendedEnterRequest wrap(
            XyzzyxJDIVirtualMachine xyzzyxvm, MonitorContendedEnterRequest req) {
        return (req == null)? null : new XyzzyxJDIMonitorContendedEnterRequest(xyzzyxvm, req);
    }

    public static XyzzyxJDIMonitorContendedEnteredRequest wrap(
            XyzzyxJDIVirtualMachine xyzzyxvm, MonitorContendedEnteredRequest req) {
        return (req == null)? null : new XyzzyxJDIMonitorContendedEnteredRequest(xyzzyxvm, req);
    }

    public static XyzzyxJDIMonitorWaitRequest wrap(
            XyzzyxJDIVirtualMachine xyzzyxvm, MonitorWaitRequest req) {
        return (req == null)? null : new XyzzyxJDIMonitorWaitRequest(xyzzyxvm, req);
    }

    public static XyzzyxJDIMonitorWaitedRequest wrap(
            XyzzyxJDIVirtualMachine xyzzyxvm, MonitorWaitedRequest req) {
        return (req == null)? null : new XyzzyxJDIMonitorWaitedRequest(xyzzyxvm, req);
    }

    public static XyzzyxJDIStepRequest wrap(
            XyzzyxJDIVirtualMachine xyzzyxvm, StepRequest req) {
        return (req == null)? null : new XyzzyxJDIStepRequest(xyzzyxvm, req);
    }

    public static XyzzyxJDIThreadDeathRequest wrap(
            XyzzyxJDIVirtualMachine xyzzyxvm, ThreadDeathRequest req) {
        return (req == null)? null : new XyzzyxJDIThreadDeathRequest(xyzzyxvm, req);
    }

    public static XyzzyxJDIThreadStartRequest wrap(
            XyzzyxJDIVirtualMachine xyzzyxvm, ThreadStartRequest req) {
        return (req == null)? null : new XyzzyxJDIThreadStartRequest(xyzzyxvm, req);
    }

    public static XyzzyxJDIVMDeathRequest wrap(XyzzyxJDIVirtualMachine xyzzyxvm, VMDeathRequest req) {
        return (req == null)? null : new XyzzyxJDIVMDeathRequest(xyzzyxvm, req);
    }

    public static List<AccessWatchpointRequest> wrapAccessWatchpointRequests(
            XyzzyxJDIVirtualMachine xyzzyxvm,  List<AccessWatchpointRequest> reqs) {
        if (reqs == null) {
            return null;
        }
         List<AccessWatchpointRequest> result = new ArrayList<>();
         for (AccessWatchpointRequest req : reqs) {
             result.add(wrap(xyzzyxvm, req));
         }
         return result;
    }

    public static List<BreakpointRequest> wrapBreakpointRequests(
            XyzzyxJDIVirtualMachine xyzzyxvm,  List<BreakpointRequest> reqs) {
        if (reqs == null) {
            return null;
        }
        List<BreakpointRequest> result = new ArrayList<>();
        for (BreakpointRequest req : reqs) {
            result.add(wrap(xyzzyxvm, req));
        }
        return result;
    }

    public static List<ClassPrepareRequest> wrapClassPrepareRequests(
            XyzzyxJDIVirtualMachine xyzzyxvm,  List<ClassPrepareRequest> reqs) {
        if (reqs == null) {
            return null;
        }
        List<ClassPrepareRequest> result = new ArrayList<>();
        for (ClassPrepareRequest req : reqs) {
            result.add(wrap(xyzzyxvm, req));
        }
        return result;
    }

    public static List<ClassUnloadRequest> wrapClassUnloadRequests(
            XyzzyxJDIVirtualMachine xyzzyxvm, List<ClassUnloadRequest> reqs) {
        if (reqs == null) {
            return null;
        }
        List<ClassUnloadRequest> result = new ArrayList<>();
        for (ClassUnloadRequest req : reqs) {
            result.add(wrap(xyzzyxvm, req));
        }
        return result;
    }

    public static List<ExceptionRequest> wrapExceptionRequests(
            XyzzyxJDIVirtualMachine xyzzyxvm, List<ExceptionRequest> reqs) {
        if (reqs == null) {
            return null;
        }
        List<ExceptionRequest> result = new ArrayList<>();
        for (ExceptionRequest req : reqs) {
            result.add(wrap(xyzzyxvm, req));
        }
        return result;
    }

    public static List<MethodEntryRequest> wrapMethodEntryRequests(
            XyzzyxJDIVirtualMachine xyzzyxvm, List<MethodEntryRequest> reqs) {
        if (reqs == null) {
            return null;
        }
        List<MethodEntryRequest> result = new ArrayList<>();
        for (MethodEntryRequest req : reqs) {
            result.add(wrap(xyzzyxvm, req));
        }
        return result;
    }

    public static List<MethodExitRequest> wrapMethodExitRequests(
            XyzzyxJDIVirtualMachine xyzzyxvm, List<MethodExitRequest> reqs) {
        if (reqs == null) {
            return null;
        }
        List<MethodExitRequest> result = new ArrayList<>();
        for (MethodExitRequest req : reqs) {
            result.add(wrap(xyzzyxvm, req));
        }
        return result;
    }

    public static List<ModificationWatchpointRequest> wrapModificationWatchpointRequests(
            XyzzyxJDIVirtualMachine xyzzyxvm, List<ModificationWatchpointRequest> reqs) {
        if (reqs == null) {
            return null;
        }
        List<ModificationWatchpointRequest> result = new ArrayList<>();
        for (ModificationWatchpointRequest req : reqs) {
            result.add(wrap(xyzzyxvm, req));
        }
        return result;
    }

    public static List<MonitorContendedEnterRequest> wrapMonitorContendedEnterRequests(
            XyzzyxJDIVirtualMachine xyzzyxvm, List<MonitorContendedEnterRequest> reqs) {
        if (reqs == null) {
            return null;
        }
        List<MonitorContendedEnterRequest> result = new ArrayList<>();
        for (MonitorContendedEnterRequest req : reqs) {
            result.add(wrap(xyzzyxvm, req));
        }
        return result;
    }

    public static List<MonitorContendedEnteredRequest> wrapMonitorContendedEnteredRequests(
            XyzzyxJDIVirtualMachine xyzzyxvm, List<MonitorContendedEnteredRequest> reqs) {
        if (reqs == null) {
            return null;
        }
        List<MonitorContendedEnteredRequest> result = new ArrayList<>();
        for (MonitorContendedEnteredRequest req : reqs) {
            result.add(wrap(xyzzyxvm, req));
        }
        return result;
    }

    public static List<MonitorWaitRequest> wrapMonitorWaitRequests(
            XyzzyxJDIVirtualMachine xyzzyxvm, List<MonitorWaitRequest> reqs) {
        if (reqs == null) {
            return null;
        }
        List<MonitorWaitRequest> result = new ArrayList<>();
        for (MonitorWaitRequest req : reqs) {
            result.add(wrap(xyzzyxvm, req));
        }
        return result;
    }

    public static List<MonitorWaitedRequest> wrapMonitorWaitedRequests(
            XyzzyxJDIVirtualMachine xyzzyxvm, List<MonitorWaitedRequest> reqs) {
        if (reqs == null) {
            return null;
        }
        List<MonitorWaitedRequest> result = new ArrayList<>();
        for (MonitorWaitedRequest req : reqs) {
            result.add(wrap(xyzzyxvm, req));
        }
        return result;
    }

    public static List<StepRequest> wrapStepRequests(
            XyzzyxJDIVirtualMachine xyzzyxvm, List<StepRequest> reqs) {
        if (reqs == null) {
            return null;
        }
        List<StepRequest> result = new ArrayList<>();
        for (StepRequest req : reqs) {
            result.add(wrap(xyzzyxvm, req));
        }
        return result;
    }

    public static List<ThreadDeathRequest> wrapThreadDeathRequests(
            XyzzyxJDIVirtualMachine xyzzyxvm, List<ThreadDeathRequest> reqs) {
        if (reqs == null) {
            return null;
        }
        List<ThreadDeathRequest> result = new ArrayList<>();
        for (ThreadDeathRequest req : reqs) {
            result.add(wrap(xyzzyxvm, req));
        }
        return result;
    }

    public static List<ThreadStartRequest> wrapThreadStartRequests(
            XyzzyxJDIVirtualMachine xyzzyxvm, List<ThreadStartRequest> reqs) {
        if (reqs == null) {
            return null;
        }
        List<ThreadStartRequest> result = new ArrayList<>();
        for (ThreadStartRequest req : reqs) {
            result.add(wrap(xyzzyxvm, req));
        }
        return result;
    }

    public static List<VMDeathRequest> wrapVMDeathRequests(
            XyzzyxJDIVirtualMachine xyzzyxvm, List<VMDeathRequest> reqs) {
        if (reqs == null) {
            return null;
        }
        List<VMDeathRequest> result = new ArrayList<>();
        for (VMDeathRequest req : reqs) {
            result.add(wrap(xyzzyxvm, req));
        }
        return result;
    }

    // unwrap methods
    public static EventRequest unwrap(Object req) {
        return (req instanceof XyzzyxJDIEventRequest)? ((XyzzyxJDIEventRequest)req).underlying() : (EventRequest)req;
    }

    public static List<EventRequest> unwrapEventRequests(List<? extends EventRequest> requests) {
        List<EventRequest> result = new ArrayList<>();
        for (EventRequest req : requests) {
            result.add((EventRequest)unwrap(req));
        }
        return result;
    }
}