/*
 * Copyright 1998-2008 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 */

package com.oracle.xyzzyx.tools.debug.tty;

import com.oracle.xyzzyx.jdi.XyzzyxJDIBootstrap;

import com.sun.jdi.*;
import com.sun.jdi.event.*;
import com.sun.jdi.request.*;
import com.sun.jdi.connect.*;

import java.util.*;
import java.io.*;

/**
 * This is a command line debugger for Xyzzyx programs. This class is same as
 * the main class of "jdb" tool except that this one uses Xyzzyx-JDI launching
 * connector by default.
 *
 */

public class TTY implements EventNotifier {
    private final Env env;
    private EventHandler handler = null;

    /**
     * List of Strings to execute at each stop.
     */
    private List<String> monitorCommands = new ArrayList<String>();
    private int monitorCount = 0;

    /**
     * The name of this tool.
     */
    private static final String progname = "jsdb";

    @Override
    public boolean shouldRemoveListener() {
        return false;
    }

    @Override
    public void vmStartEvent(VMStartEvent se)  {
        Thread.yield();  // fetch output
        env.messageOutput().lnprint("VM Started:");
    }

    @Override
    public void vmDeathEvent(VMDeathEvent e)  {
    }

    @Override
    public void vmDisconnectEvent(VMDisconnectEvent e)  {
    }

    @Override
    public void threadStartEvent(ThreadStartEvent e)  {
    }

    @Override
    public void threadDeathEvent(ThreadDeathEvent e)  {
    }

    @Override
    public void classPrepareEvent(ClassPrepareEvent e)  {
    }

    @Override
    public void classUnloadEvent(ClassUnloadEvent e)  {
    }

    @Override
    public void breakpointEvent(BreakpointEvent be)  {
        Thread.yield();  // fetch output
        env.messageOutput().lnprint("Breakpoint hit:");
    }

    @Override
    public void fieldWatchEvent(WatchpointEvent fwe)  {
        Field field = fwe.field();
        ObjectReference obj = fwe.object();
        Thread.yield();  // fetch output

        if (fwe instanceof ModificationWatchpointEvent) {
            env.messageOutput().lnprint("Field access encountered before after",
                                  new Object [] {field,
                                                 fwe.valueCurrent(),
                                                 ((ModificationWatchpointEvent)fwe).valueToBe()});
        } else {
            env.messageOutput().lnprint("Field access encountered", field.toString());
        }
    }

    @Override
    public void stepEvent(StepEvent se)  {
        Thread.yield();  // fetch output
        env.messageOutput().lnprint("Step completed:");
    }

    @Override
    public void exceptionEvent(ExceptionEvent ee) {
        Thread.yield();  // fetch output
        Location catchLocation = ee.catchLocation();
        if (catchLocation == null) {
            env.messageOutput().lnprint("Exception occurred uncaught",
                                  ee.exception().referenceType().name());
        } else {
            env.messageOutput().lnprint("Exception occurred caught",
                                  new Object [] {ee.exception().referenceType().name(),
                                                 Commands.locationString(catchLocation)});
        }
    }

    @Override
    public void methodEntryEvent(MethodEntryEvent me) {
        Thread.yield();  // fetch output
        /*
         * These can be very numerous, so be as efficient as possible.
         * If we are stopping here, then we will see the normal location
         * info printed.
         */
        if (me.request().suspendPolicy() != EventRequest.SUSPEND_NONE) {
            // We are stopping; the name will be shown by the normal mechanism
            env.messageOutput().lnprint("Method entered:");
        } else {
            // We aren't stopping, show the name
            env.messageOutput().print("Method entered:");
            printLocationOfEvent(me);
        }
    }

    @Override
    public boolean methodExitEvent(MethodExitEvent me) {
        Thread.yield();  // fetch output
        /*
         * These can be very numerous, so be as efficient as possible.
         */
        Method mmm = env.atExitMethod();
        Method meMethod = me.method();

        if (mmm == null || mmm.equals(meMethod)) {
            // Either we are not tracing a specific method, or we are
            // and we are exitting that method.

            if (me.request().suspendPolicy() != EventRequest.SUSPEND_NONE) {
                // We will be stopping here, so do a newline
                env.messageOutput().println();
            }
            if (env.vm().canGetMethodReturnValues()) {
                env.messageOutput().print("Method exitedValue:", me.returnValue() + "");
            } else {
                env.messageOutput().print("Method exited:");
            }

            if (me.request().suspendPolicy() == EventRequest.SUSPEND_NONE) {
                // We won't be stopping here, so show the method name
                printLocationOfEvent(me);

            }

            // In case we want to have a one shot trace exit some day, this
            // code disables the request so we don't hit it again.
            if (false) {
                // This is a one shot deal; we don't want to stop
                // here the next time.
                env.setAtExitMethod(null);
                EventRequestManager erm = env.vm().eventRequestManager();
                for (EventRequest eReq : erm.methodExitRequests()) {
                    if (eReq.equals(me.request())) {
                        eReq.disable();
                    }
                }
            }
            return true;
        }

        // We are tracing a specific method, and this isn't it.  Keep going.
        return false;
    }

    @SuppressWarnings("CallToThreadYield")
    @Override
    public void vmInterrupted() {
        Thread.yield();  // fetch output
        printCurrentLocation();
        for (String cmd : monitorCommands) {
            StringTokenizer t = new StringTokenizer(cmd);
            t.nextToken();  // get rid of monitor number
            executeCommand(t);
        }
        env.printPrompt();
    }

    @Override
    public void receivedEvent(Event event) {
    }

    private void printBaseLocation(String threadName, Location loc) {
        env.messageOutput().println("location",
                              new Object [] {threadName,
                                             Commands.locationString(loc)});
    }

    private void printCurrentLocation() {
        ThreadInfo threadInfo = env.getCurrentThreadInfo();
        StackFrame frame;
        try {
            frame = threadInfo.getCurrentFrame();
        } catch (IncompatibleThreadStateException exc) {
            env.messageOutput().println("<location unavailable>");
            return;
        }
        if (frame == null) {
            env.messageOutput().println("No frames on the current call stack");
        } else {
            Location loc = frame.location();
            printBaseLocation(threadInfo.getThread().name(), loc);
            // Output the current source line, if possible
            if (loc.lineNumber() != -1) {
                String line;
                try {
                    line = env.sourceLine(loc, loc.lineNumber());
                } catch (java.io.IOException e) {
                    line = null;
                }
                if (line != null) {
                    env.messageOutput().println("source line number and line",
                                          new Object [] {new Integer(loc.lineNumber()),
                                                         line});
                }
            }
        }
        env.messageOutput().println();
    }

    private void printLocationOfEvent(LocatableEvent theEvent) {
        printBaseLocation(theEvent.thread().name(), theEvent.location());
    }

    void help() {
        env.messageOutput().println("zz help text");
    }

    private static final String[][] commandList = {
        /*
         * NOTE: this list must be kept sorted in ascending ASCII
         *       order by element [0].  Ref: isCommand() below.
         *
         *Command      OK when        OK when
         * name      disconnected?   readonly?
         *------------------------------------
         */
        {"!!",           "n",         "y"},
        {"?",            "y",         "y"},
        {"bytecodes",    "n",         "y"},
        {"catch",        "y",         "n"},
        {"class",        "n",         "y"},
        {"classes",      "n",         "y"},
        {"classpath",    "n",         "y"},
        {"clear",        "y",         "n"},
        {"connectors",   "y",         "y"},
        {"cont",         "n",         "n"},
        {"disablegc",    "n",         "n"},
        {"down",         "n",         "y"},
        {"dump",         "n",         "y"},
        {"enablegc",     "n",         "n"},
        {"eval",         "n",         "y"},
        {"exclude",      "y",         "n"},
        {"exit",         "y",         "y"},
        {"extension",    "n",         "y"},
        {"fields",       "n",         "y"},
        {"gc",           "n",         "n"},
        {"help",         "y",         "y"},
        {"ignore",       "y",         "n"},
        {"interrupt",    "n",         "n"},
        {"kill",         "n",         "n"},
        {"lines",        "n",         "y"},
        {"list",         "n",         "y"},
        {"load",         "n",         "y"},
        {"locals",       "n",         "y"},
        {"lock",         "n",         "n"},
        {"memory",       "n",         "y"},
        {"methods",      "n",         "y"},
        {"monitor",      "n",         "n"},
        {"next",         "n",         "n"},
        {"pop",          "n",         "n"},
        {"print",        "n",         "y"},
        {"quit",         "y",         "y"},
        {"read",         "y",         "y"},
        {"redefine",     "n",         "n"},
        {"reenter",      "n",         "n"},
        {"resume",       "n",         "n"},
        {"run",          "y",         "n"},
        {"save",         "n",         "n"},
        {"set",          "n",         "n"},
        {"sourcepath",   "y",         "y"},
        {"step",         "n",         "n"},
        {"stepi",        "n",         "n"},
        {"stop",         "y",         "n"},
        {"suspend",      "n",         "n"},
        {"thread",       "n",         "y"},
        {"threadgroup",  "n",         "y"},
        {"threadgroups", "n",         "y"},
        {"threadlocks",  "n",         "y"},
        {"threads",      "n",         "y"},
        {"trace",        "n",         "n"},
        {"unmonitor",    "n",         "n"},
        {"untrace",      "n",         "n"},
        {"unwatch",      "y",         "n"},
        {"up",           "n",         "y"},
        {"use",          "y",         "y"},
        {"version",      "y",         "y"},
        {"watch",        "y",         "n"},
        {"where",        "n",         "y"},
        {"wherei",       "n",         "y"},
    };

    /*
     * Look up the command string in commandList.
     * If found, return the index.
     * If not found, return index < 0
     */
    private int isCommand(String key) {
        //Reference: binarySearch() in java/util/Arrays.java
        //           Adapted for use with String[][0].
        int low = 0;
        int high = commandList.length - 1;
        long i = 0;
        while (low <= high) {
            int mid = (low + high) >>> 1;
            String midVal = commandList[mid][0];
            int compare = midVal.compareTo(key);
            if (compare < 0)
                low = mid + 1;
            else if (compare > 0)
                high = mid - 1;
            else
                return mid; // key found
        }
        return -(low + 1);  // key not found.
    };

    /*
     * Return true if the command is OK when disconnected.
     */
    private boolean isDisconnectCmd(int ii) {
        if (ii < 0 || ii >= commandList.length) return false;
        return (commandList[ii][1].equals("y"));
    }

    /*
     * Return true if the command is OK when readonly.
     */
    private boolean isReadOnlyCmd(int ii) {
        if (ii < 0 || ii >= commandList.length) return false;
        return (commandList[ii][2].equals("y"));
    };


    void executeCommand(StringTokenizer t) {
        String cmd = t.nextToken().toLowerCase();
        // Normally, prompt for the next command after this one is done
        boolean showPrompt = true;


        /*
         * Anything starting with # is discarded as a no-op or 'comment'.
         */
        if (!cmd.startsWith("#")) {
            /*
             * Next check for an integer repetition prefix.  If found,
             * recursively execute cmd that number of times.
             */
            if (Character.isDigit(cmd.charAt(0)) && t.hasMoreTokens()) {
                try {
                    int repeat = Integer.parseInt(cmd);
                    String subcom = t.nextToken("");
                    while (repeat-- > 0) {
                        executeCommand(new StringTokenizer(subcom));
                        showPrompt = false; // Bypass the printPrompt() below.
                    }
                } catch (NumberFormatException exc) {
                    env.messageOutput().println("Unrecognized command.  Try help...", cmd);
                }
            } else {
                int commandNumber = isCommand(cmd);
                /*
                 * Check for an unknown command
                 */
                if (commandNumber < 0) {
                    env.messageOutput().println("Unrecognized command.  Try help...", cmd);
                } else if (!env.connection().isOpen() && !isDisconnectCmd(commandNumber)) {
                    env.messageOutput().println("Command not valid until the VM is started with the run command",
                                          cmd);
                } else if (env.connection().isOpen() && !env.vm().canBeModified() &&
                           !isReadOnlyCmd(commandNumber)) {
                    env.messageOutput().println("Command is not supported on a read-only VM connection",
                                          cmd);
                } else {

                    Commands evaluator = new Commands(env);
                    try {
                        switch (cmd) {
                            case "print":
                                evaluator.commandPrint(t, false);
                                showPrompt = false;        // asynchronous command
                                break;
                            case "eval":
                                evaluator.commandPrint(t, false);
                                showPrompt = false;        // asynchronous command
                                break;
                            case "set":
                                evaluator.commandSet(t);
                                showPrompt = false;        // asynchronous command
                                break;
                            case "dump":
                                evaluator.commandPrint(t, true);
                                showPrompt = false;        // asynchronous command
                                break;
                            case "locals":
                                evaluator.commandLocals();
                                break;
                            case "classes":
                                evaluator.commandClasses();
                                break;
                            case "class":
                                evaluator.commandClass(t);
                                break;
                            case "connectors":
                                evaluator.commandConnectors(XyzzyxJDIBootstrap.virtualMachineManager());
                                break;
                            case "methods":
                                evaluator.commandMethods(t);
                                break;
                            case "fields":
                                evaluator.commandFields(t);
                                break;
                            case "threads":
                                evaluator.commandThreads(t);
                                break;
                            case "thread":
                                evaluator.commandThread(t);
                                break;
                            case "suspend":
                                evaluator.commandSuspend(t);
                                break;
                            case "resume":
                                evaluator.commandResume(t);
                                break;
                            case "cont":
                                evaluator.commandCont();
                                break;
                            case "threadgroups":
                                evaluator.commandThreadGroups();
                                break;
                            case "threadgroup":
                                evaluator.commandThreadGroup(t);
                                break;
                            case "catch":
                                evaluator.commandCatchException(t);
                                break;
                            case "ignore":
                                evaluator.commandIgnoreException(t);
                                break;
                            case "step":
                                evaluator.commandStep(t);
                                break;
                            case "stepi":
                                evaluator.commandStepi();
                                break;
                            case "next":
                                evaluator.commandNext();
                                break;
                            case "kill":
                                evaluator.commandKill(t);
                                break;
                            case "interrupt":
                                evaluator.commandInterrupt(t);
                                break;
                            case "trace":
                                evaluator.commandTrace(t);
                                break;
                            case "untrace":
                                evaluator.commandUntrace(t);
                                break;
                            case "where":
                                evaluator.commandWhere(t, false);
                                break;
                            case "wherei":
                                evaluator.commandWhere(t, true);
                                break;
                            case "up":
                                evaluator.commandUp(t);
                                break;
                            case "down":
                                evaluator.commandDown(t);
                                break;
                            case "load":
                                evaluator.commandLoad(t);
                                break;
                            case "run":
                                evaluator.commandRun(t);
                                /*
                                 * Fire up an event handler, if the connection was just
                                 * opened. Since this was done from the run command
                                 * we don't stop the VM on its VM start event (so
                                 * arg 2 is false).
                                 */
                                if ((handler == null) && env.connection().isOpen()) {
                                    handler = new EventHandler(env, this, false);
                                }
                                break;
                            case "memory":
                                evaluator.commandMemory();
                                break;
                            case "gc":
                                evaluator.commandGC();
                                break;
                            case "stop":
                                evaluator.commandStop(t);
                                break;
                            case "clear":
                                evaluator.commandClear(t);
                                break;
                            case "watch":
                                evaluator.commandWatch(t);
                                break;
                            case "unwatch":
                                evaluator.commandUnwatch(t);
                                break;
                            case "list":
                                evaluator.commandList(t);
                                break;
                            case "lines":
                                // Undocumented command: useful for testing.
                                evaluator.commandLines(t);
                                break;
                            case "classpath":
                                evaluator.commandClasspath(t);
                                break;
                            case "use":
                            case "sourcepath":
                                evaluator.commandUse(t);
                                break;
                            case "monitor":
                                monitorCommand(t);
                                break;
                            case "unmonitor":
                                unmonitorCommand(t);
                                break;
                            case "lock":
                                evaluator.commandLock(t);
                                showPrompt = false;        // asynchronous command
                                break;
                            case "threadlocks":
                                evaluator.commandThreadlocks(t);
                                break;
                            case "disablegc":
                                evaluator.commandDisableGC(t);
                                showPrompt = false;        // asynchronous command
                                break;
                            case "enablegc":
                                evaluator.commandEnableGC(t);
                                showPrompt = false;        // asynchronous command
                                break;
                            case "save":
                                // Undocumented command: useful for testing.
                                evaluator.commandSave(t);
                                showPrompt = false;        // asynchronous command
                                break;
                            case "bytecodes":
                                // Undocumented command: useful for testing.
                                evaluator.commandBytecodes(t);
                                break;
                            case "redefine":
                                evaluator.commandRedefine(t);
                                break;
                            case "pop":
                                evaluator.commandPopFrames(t, false);
                                break;
                            case "reenter":
                                evaluator.commandPopFrames(t, true);
                                break;
                            case "extension":
                                evaluator.commandExtension(t);
                                break;
                            case "exclude":
                                evaluator.commandExclude(t);
                                break;
                            case "read":
                                readCommand(t);
                                break;
                            case "help":
                            case "?":
                                help();
                                break;
                            case "version":
                                evaluator.commandVersion(progname,
                                                         XyzzyxJDIBootstrap.virtualMachineManager());
                                break;
                            case "quit":
                            case "exit":
                                if (handler != null) {
                                    handler.shutdown();
                                }
                                env.shutdown();
                                break;
                            default:
                                env.messageOutput().println("Unrecognized command.  Try help...", cmd);
                                break;
                        }
                    } catch (VMCannotBeModifiedException rovm) {
                        env.messageOutput().println("Command is not supported on a read-only VM connection", cmd);
                    } catch (UnsupportedOperationException uoe) {
                        env.messageOutput().println("Command is not supported on the target VM", cmd);
                    } catch (VMNotConnectedException vmnse) {
                        env.messageOutput().println("Command not valid until the VM is started with the run command",
                                              cmd);
                    } catch (Exception e) {
                        env.messageOutput().printException("Internal exception:", e);
                    }
                }
            }
        }
        if (showPrompt) {
            env.printPrompt();
        }
    }

    /*
     * Maintain a list of commands to execute each time the VM is suspended.
     */
    void monitorCommand(StringTokenizer t) {
        if (t.hasMoreTokens()) {
            ++monitorCount;
            monitorCommands.add(monitorCount + ": " + t.nextToken(""));
        } else {
            for (String cmd : monitorCommands) {
                env.messageOutput().printDirectln(cmd);// Special case: use printDirectln()
            }
        }
    }

    void unmonitorCommand(StringTokenizer t) {
        if (t.hasMoreTokens()) {
            String monTok = t.nextToken();
            int monNum;
            try {
                monNum = Integer.parseInt(monTok);
            } catch (NumberFormatException exc) {
                env.messageOutput().println("Not a monitor number:", monTok);
                return;
            }
            String monStr = monTok + ":";
            for (String cmd : monitorCommands) {
                StringTokenizer ct = new StringTokenizer(cmd);
                if (ct.nextToken().equals(monStr)) {
                    monitorCommands.remove(cmd);
                    env.messageOutput().println("Unmonitoring", cmd);
                    return;
                }
            }
            env.messageOutput().println("No monitor numbered:", monTok);
        } else {
            env.messageOutput().println("Usage: unmonitor <monitor#>");
        }
    }


    void readCommand(StringTokenizer t) {
        if (t.hasMoreTokens()) {
            String cmdfname = t.nextToken();
            if (!readCommandFile(new File(cmdfname))) {
                env.messageOutput().println("Could not open:", cmdfname);
            }
        } else {
            env.messageOutput().println("Usage: read <command-filename>");
        }
    }

    /**
     * Read and execute a command file.  Return true if the file was read
     * else false;
     */
    boolean readCommandFile(File f) {
        BufferedReader inFile = null;
        try {
            if (f.canRead()) {
                // Process initial commands.
                env.messageOutput().println("*** Reading commands from", f.getPath());
                inFile = new BufferedReader(new FileReader(f));
                String ln;
                while ((ln = inFile.readLine()) != null) {
                    StringTokenizer t = new StringTokenizer(ln);
                    if (t.hasMoreTokens()) {
                        executeCommand(t);
                    }
                }
            }
        } catch (IOException e) {
        } finally {
            if (inFile != null) {
                try {
                    inFile.close();
                } catch (Exception exc) {
                }
            }
        }
        return inFile != null;
    }

    /**
     * Try to read commands from dir/fname, unless
     * the canonical path passed in is the same as that
     * for dir/fname.
     * Return null if that file doesn't exist,
     * else return the canonical path of that file.
     */
    String readStartupCommandFile(String dir, String fname, String canonPath) {
        File dotInitFile = new File(dir, fname);
        if (!dotInitFile.exists()) {
            return null;
        }

        String myCanonFile;
        try {
            myCanonFile = dotInitFile.getCanonicalPath();
        } catch (IOException ee) {
            env.messageOutput().println("Could not open:", dotInitFile.getPath());
            return null;
        }
        if (canonPath == null || !canonPath.equals(myCanonFile)) {
            if (!readCommandFile(dotInitFile)) {
                env.messageOutput().println("Could not open:", dotInitFile.getPath());
            }
        }
        return myCanonFile;
    }

    TTY(Env env) throws Exception {

        this.env = env;
        env.messageOutput().println("Initializing progname", progname);

        if (env.connection().isOpen() && env.vm().canBeModified()) {
            /*
             * Connection opened on startup. Start event handler
             * immediately, telling it (through arg 2) to stop on the
             * VM start event.
             */
            this.handler = new EventHandler(env, this, true);
        }
        try {
            BufferedReader in =
                    new BufferedReader(new InputStreamReader(System.in));

            String lastLine = null;

            Thread.currentThread().setPriority(Thread.NORM_PRIORITY);

            /*
             * Read start up files.  This mimics the behavior
             * of gdb which will read both ~/.gdbinit and then
             * ./.gdbinit if they exist.  We have the twist that
             * we allow two different names, so we do this:
             *  if ~/jdb.ini exists,
             *      read it
             *  else if ~/.jdbrc exists,
             *      read it
             *
             *  if ./jdb.ini exists,
             *      if it hasn't been read, read it
             *      It could have been read above because ~ == .
             *      or because of symlinks, ...
             *  else if ./jdbrx exists
             *      if it hasn't been read, read it
             */
            {
                String userHome = System.getProperty("user.home");
                String canonPath;

                if ((canonPath = readStartupCommandFile(userHome, "jdb.ini", null)) == null) {
                    // Doesn't exist, try alternate spelling
                    canonPath = readStartupCommandFile(userHome, ".jdbrc", null);
                }

                String userDir = System.getProperty("user.dir");
                if (readStartupCommandFile(userDir, "jdb.ini", canonPath) == null) {
                    // Doesn't exist, try alternate spelling
                    readStartupCommandFile(userDir, ".jdbrc", canonPath);
                }
            }

            // Process interactive commands.
            env.printPrompt();
            while (true) {
                String ln = in.readLine();
                if (ln == null) {
                    env.messageOutput().println("Input stream closed.");
                    ln = "quit";
                }

                if (ln.startsWith("!!") && lastLine != null) {
                    ln = lastLine + ln.substring(2);
                    env.messageOutput().printDirectln(ln);// Special case: use printDirectln()
                }

                StringTokenizer t = new StringTokenizer(ln);
                if (t.hasMoreTokens()) {
                    lastLine = ln;
                    executeCommand(t);
                } else {
                    env.printPrompt();
                }
            }
        } catch (VMDisconnectedException e) {
            handler.handleDisconnectedException();
        }
    }

    private static void usage(Env env) {
        env.messageOutput().println("zz usage text", new Object [] {progname,
                                                     File.pathSeparator});
        System.exit(1);
    }

    static void usageError(Env env, String messageKey) {
        env.messageOutput().println(messageKey);
        env.messageOutput().println();
        usage(env);
    }

    static void usageError(Env env, String messageKey, String argument) {
        env.messageOutput().println(messageKey, argument);
        env.messageOutput().println();
        usage(env);
    }

    private static boolean supportsSharedMemory() {
        for (Connector connector :
                 XyzzyxJDIBootstrap.virtualMachineManager().allConnectors()) {
            if (connector.transport() == null) {
                continue;
            }
            if ("dt_shmem".equals(connector.transport().name())) {
                return true;
            }
        }
        return false;
    }

    private static String addressToSocketArgs(String address) {
        int index = address.indexOf(':');
        if (index != -1) {
            String hostString = address.substring(0, index);
            String portString = address.substring(index + 1);
            return "hostname=" + hostString + ",port=" + portString;
        } else {
            return "port=" + address;
        }
    }

    private static boolean hasWhitespace(String string) {
        int length = string.length();
        for (int i = 0; i < length; i++) {
            if (Character.isWhitespace(string.charAt(i))) {
                return true;
            }
        }
        return false;
    }

    private static String addArgument(String string, String argument) {
        if (hasWhitespace(argument) || argument.indexOf(',') != -1) {
            // Quotes were stripped out for this argument, add 'em back.
            StringBuilder buffer = new StringBuilder(string);
            buffer.append('"');
            for (int i = 0; i < argument.length(); i++) {
                char c = argument.charAt(i);
                if (c == '"') {
                    buffer.append('\\');
                }
                buffer.append(c);
            }
            buffer.append("\" ");
            return buffer.toString();
        } else {
            return string + argument + ' ';
        }
    }

    @SuppressWarnings("ResultOfObjectAllocationIgnored")
    public static void main(String argv[]) throws MissingResourceException {
        String cmdLine = "";
        String javaArgs = "";
        int traceFlags = VirtualMachine.TRACE_NONE;
        boolean launchImmediately = false;
        String connectSpec = null;
        Env env = new Env();

        for (int i = 0; i < argv.length; i++) {
            String token = argv[i];
            if (token.equals("-dbgtrace")) {
                if ((i == argv.length - 1) ||
                    ! Character.isDigit(argv[i+1].charAt(0))) {
                    traceFlags = VirtualMachine.TRACE_ALL;
                } else {
                    String flagStr = "";
                    try {
                        flagStr = argv[++i];
                        traceFlags = Integer.decode(flagStr).intValue();
                    } catch (NumberFormatException nfe) {
                        usageError(env, "dbgtrace flag value must be an integer:",
                                   flagStr);
                        return;
                    }
                }
            } else if (token.equals("-X")) {
                usageError(env,  "Use java minus X to see");
                return;
            } else if (
                   // Standard VM options passed on
                   token.equals("-v") || token.startsWith("-v:") ||  // -v[:...]
                   token.startsWith("-verbose") ||                  // -verbose[:...]
                   token.startsWith("-D") ||
                   // -classpath handled below
                   // NonStandard options passed on
                   token.startsWith("-X") ||
                   // Old-style options (These should remain in place as long as
                   //  the standard VM accepts them)
                   token.equals("-noasyncgc") || token.equals("-prof") ||
                   token.equals("-verify") || token.equals("-noverify") ||
                   token.equals("-verifyremote") ||
                   token.equals("-verbosegc") ||
                   token.startsWith("-ms") || token.startsWith("-mx") ||
                   token.startsWith("-ss") || token.startsWith("-oss") ) {

                javaArgs = addArgument(javaArgs, token);
            } else if (token.equals("-tclassic")) {
                usageError(env, "Classic VM no longer supported.");
                return;
            } else if (token.equals("-tclient")) {
                // -client must be the first one
                javaArgs = "-client " + javaArgs;
            } else if (token.equals("-tserver")) {
                // -server must be the first one
                javaArgs = "-server " + javaArgs;
            } else if (token.equals("-sourcepath")) {
                if (i == (argv.length - 1)) {
                    usageError(env, "No sourcepath specified.");
                    return;
                }
                env.setSourcePath(argv[++i]);
            } else if (token.equals("-classpath")) {
                if (i == (argv.length - 1)) {
                    usageError(env, "No classpath specified.");
                    return;
                }
                javaArgs = addArgument(javaArgs, token);
                javaArgs = addArgument(javaArgs, argv[++i]);
            } else if (token.equals("-attach")) {
                if (connectSpec != null) {
                    usageError(env, "cannot redefine existing connection", token);
                    return;
                }
                if (i == (argv.length - 1)) {
                    usageError(env, "No attach address specified.");
                    return;
                }
                String address = argv[++i];

                /*
                 * -attach is shorthand for one of the Xyzzyx-JDI implementation's
                 * attaching connectors. Use the shared memory attach if it's
                 * available; otherwise, use sockets. Build a connect
                 * specification string based on this decision.
                 */
                if (supportsSharedMemory()) {
                    connectSpec = "com.oracle.xyzzyx.jdi.connect.XyzzyxJDISharedMemoryAttachingConnector:name=" +
                                   address;
                } else {
                    String suboptions = addressToSocketArgs(address);
                    connectSpec = "com.oracle.xyzzyx.jdi.connect.XyzzyxJDISocketAttachingConnector:" + suboptions;
                }
            } else if (token.equals("-listen") || token.equals("-listenany")) {
                if (connectSpec != null) {
                    usageError(env, "cannot redefine existing connection", token);
                    return;
                }
                String address = null;
                if (token.equals("-listen")) {
                    if (i == (argv.length - 1)) {
                        usageError(env, "No attach address specified.");
                        return;
                    }
                    address = argv[++i];
                }

                /*
                 * -listen[any] is shorthand for one of the Xyzzyx-JDI implementation's
                 * listening connectors. Use the shared memory listen if it's
                 * available; otherwise, use sockets. Build a connect
                 * specification string based on this decision.
                 */
                if (supportsSharedMemory()) {
                    connectSpec = "com.oracle.xyzzyx.jdi.connect.XyzzyxJDISharedMemoryListeningConnector:";
                    if (address != null) {
                        connectSpec += ("name=" + address);
                    }
                } else {
                    connectSpec = "com.oracle.xyzzyx.jdi.connect.XyzzyxJDISocketListeningConnector:";
                    if (address != null) {
                        connectSpec += addressToSocketArgs(address);
                    }
                }
            } else if (token.equals("-launch")) {
                launchImmediately = true;
            } else if (token.equals("-listconnectors")) {
                Commands evaluator = new Commands(env);
                evaluator.commandConnectors(XyzzyxJDIBootstrap.virtualMachineManager());
                return;
            } else if (token.equals("-connect")) {
                /*
                 * -connect allows the user to pick the connector
                 * used in bringing up the target VM. This allows
                 * use of connectors other than those in the reference
                 * implementation.
                 */
                if (connectSpec != null) {
                    usageError(env, "cannot redefine existing connection", token);
                    return;
                }
                if (i == (argv.length - 1)) {
                    usageError(env, "No connect specification.");
                    return;
                }
                connectSpec = argv[++i];
            } else if (token.equals("-help")) {
                usage(env);
            } else if (token.equals("-version")) {
                Commands evaluator = new Commands(env);
                evaluator.commandVersion(progname,
                                         XyzzyxJDIBootstrap.virtualMachineManager());
                System.exit(0);
            } else if (token.startsWith("-")) {
                usageError(env, "invalid option", token);
                return;
            } else {
                // Everything from here is part of the command line
                cmdLine = addArgument("", token);
                for (i++; i < argv.length; i++) {
                    cmdLine = addArgument(cmdLine, argv[i]);
                }
                break;
            }
        }

        /*
         * Unless otherwise specified, set the default connect spec.
         */

        /*
         * Here are examples of jsdb command lines and how the options
         * are interpreted as arguments to the program being debugged.
         * arg1       arg2
         * ----       ----
         * jsdb hello a b       a          b
         * jsdb hello "a b"     a b
         * jsdb hello a,b       a,b
         * jsdb hello a, b      a,         b
         * jsdb hello "a, b"    a, b
         * jsdb -connect "com.sun.jdi.CommandLineLaunch:main=hello  a,b"   illegal
         * jsdb -connect  com.sun.jdi.CommandLineLaunch:main=hello "a,b"   illegal
         * jsdb -connect 'com.sun.jdi.CommandLineLaunch:main=hello "a,b"'  arg1 = a,b
         * jsdb -connect 'com.sun.jdi.CommandLineLaunch:main=hello "a b"'  arg1 = a b
         * jsdb -connect 'com.sun.jdi.CommandLineLaunch:main=hello  a b'   arg1 = a  arg2 = b
         * jsdb -connect 'com.sun.jdi.CommandLineLaunch:main=hello "a," b' arg1 = a, arg2 = b
         */
        if (connectSpec == null) {
            connectSpec = "com.oracle.xyzzyx.jdi.connect.XyzzyxJDILaunchingConnector:";
        } else if (!connectSpec.endsWith(",") && !connectSpec.endsWith(":")) {
            connectSpec += ","; // (Bug ID 4285874)
        }

        cmdLine = cmdLine.trim();
        javaArgs = javaArgs.trim();

        if (cmdLine.length() > 0) {
            if (!connectSpec.startsWith("com.oracle.xyzzyx.jdi.connect.XyzzyxJDILaunchingConnector:") &&
                !connectSpec.startsWith("com.sun.jdi.CommandLineLaunch:")) {
                usageError(env, "Cannot specify command line with connector:",
                           connectSpec);
                return;
            }
            connectSpec += "main=" + cmdLine + ",";
        }

        if (javaArgs.length() > 0) {
            if (!connectSpec.startsWith("com.oracle.xyzzyx.jdi.connect.XyzzyxJDILaunchingConnector:") &&
                !connectSpec.startsWith("com.sun.jdi.CommandLineLaunch:")) {
                usageError(env, "Cannot specify target vm arguments with connector:",
                           connectSpec);
                return;
            }
            connectSpec += "options=" + javaArgs + ",";
        }

        try {
            if (! connectSpec.endsWith(",")) {
                connectSpec += ","; // (Bug ID 4285874)
            }
            env.init(connectSpec, launchImmediately, traceFlags);
            new TTY(env);
        } catch(Exception e) {
            env.messageOutput().printException("Internal exception:", e);
        }
    }
}
